# Feb 15. Change OLS to a weighted LS.
import time as tm
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import math
import scipy
from scipy import stats
import sys
import sklearn.preprocessing
from sklearn.preprocessing import MinMaxScaler
from scipy import stats
import scipy.stats as stats
from scipy.optimize import nnls
from scipy import linalg
from sklearn.linear_model import LinearRegression
from sklearn.externals.joblib import Parallel, delayed
#################################################################
#
# The R matrix on INPUT is expected to contain the covariates as well.
# The R matrix is usually wither the original RAW R data OR the df_new_R
# data object form running this Class a prior iteration. The best choice is TBD
# Then the final imputted R wil contain the same covariates
# The format of the NN matrices from the preceding Clustering class is
# CELL NN1 NN2 .... NNk+1
# The values can either be the indexing OR the distances depending on the file you are reading 
# Feb 15. Change OLS to a weighted LS.
# NOTE: WLS yield BLUE if weights are inverse variances

# CHECK if the reshape(-1,numGenes is correct)

# TODO read batchfilename as an option for covariate correction

# Might consider usingh KNeighborsRegressor directly !
import time as tm
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import math 
import scipy 
from scipy import stats
import sys
import sklearn.preprocessing
from sklearn.preprocessing import MinMaxScaler
from scipy import stats
import scipy.stats as stats
from scipy import linalg
from sklearn.linear_model import LinearRegression
from sklearn.externals.joblib import Parallel, delayed

#########################################################################
# Keep the known covariates of input R

def getKnownGroupEffects(fulldata,batchfilename):
    """ Check the input file for KNOWN possible group effects
    Exclude those that may not exist such as TREATMENT. Make no
    assumptions about order or sorting. Return list in arbitrary 
    order. Independent of list provided by the user
    """
    foundColumns = list()
    listOneHotEncoded = list()
    if (batchfilename == False):
        print('Using prechosen batch effect names')
        currentPossibleColumnEffects = ['TREATMENT','MOUSE','SEX','DATE','NODEID']
        allColumns = fulldata.columns.values
        listOneHotEncoded = ['MOUSE'] # By default for use with our data testing
        # Grab only those fulldata columns that exist inb currentPossible
        foundColumns = list(set(allColumns) & set(currentPossibleColumnEffects))
    else:
        foundColumnsCombined = list()
        with open(batchfilename) as f:
            tempColumns = f.read().splitlines()
            for effect in tempColumns:
                #foundColumnsCombined.append(effect.upper())
                foundColumnsCombined.append(effect)
            for ipair in foundColumnsCombined:
                #print('effect is '+ipair)
                name,typeEnc = ipair.strip().split(' ')
                foundColumns.append(name)
                if (typeEnc=='ONEHOT'):
                    listOneHotEncoded.append(name)
    print('Number of Possible batch effects found/specified is '+str(len(foundColumns)))
    print('Number of groupeffect columns in data is '+str(foundColumns))
    #print('They have been converted to UPPER CASE: '+str(foundColumns))
    return (foundColumns,listOneHotEncoded)

def buildCellIndexMap(NNindex):
    """Build a dataframe with integer IDs as index and cell names as values
    One would prefer to have grabbed the IDs from the indices file but NOT!
    """
    cellNames=NNindex.index.values
    #cellIDs=NNindex['CELLTARGET'].values
    # Because of NN behavior this previous step doesn't always work
    # SO just give them integer values assuming input order is retained
    dataMap = pd.DataFrame(cellNames)
    ##dataMap.index = cellIDs
    dataMap.columns=['CELL']
    return (dataMap)

def aggregateNNLS(Xdata,Ydata):
    """Must loop over all features for the given icell
    """
    numObservations = Ydata.shape[1]
    w = np.zeros(numObservations)
    rnorm = np.zeros(numObservations)
    for i in range(0,numObservations):
        Y = Ydata[:,i]
        wtemp,rnormtemp = nnls(Xdata.reshape(-1,1),Y.flatten())
        w[i]=wtemp
        rnorm[i]=rnormtemp
    return(w,rnorm)

def singleNNLS(Y,X):
    """Must loop over all features for the given icell
    """
    print(str(Y))
    w,rnorm = nnls(X.reshape(-1,1),Y)
    return w


########################################################
class regressDataClass(object):
    """Using Regression,predict EVERY cell read from the input (unscaled) reads
    using the top k nearest neighbors are identified in the input INDICES object.
    For the regression we INCLUDE the target cell meaning that we actually regress k+1 terms
    On return is an unscaled new R matrix (nxp) and includes all the covariates and S which is a 
    scaled version of R.
    R is returns as reads with per-cell counts normalized
    S is returned with per-cell counts normalized, features scaled to N(0,1). and then log(read+1)

    Attributes
    ----------
    self.kNN = Sp[ecify number of nearest neighbors (Excluding self term) ( so k not k+1)
            I.e., same value as passed to the generateCluster methods
    self.batchfilename (opt) file that contains n rows and 2 column withs relevant batch effect names
            col 1 effect name: col two either ONEHOT or CAT for type of inclusion. Same file as used for project
    self.rawfilename. The input read data from which regression will proceeed. Values must be >=0 as the total 
            cell counts is determined from the file ands used for X in the OLS
    self.xtildeNeighborIndicesFilename. Input neighbor matrix (n by k+1)
    self.xtildeNeighborIndicesFilename. Input neighbor distance matrix (n by k+1) (not used at this time)
    self.outputRfilename. Output imputed R matrix
    self.outputSfilename. Output scaled version of imputed R
    self.origScaling = True feature scale as mean-centered, std=1 else MinMax (0,1)
    self.logScaledResults True perform log(R+1) transform on the data
    """

    def __init__(self, kNN, inrawfilename, indexfilename, distancefilename, outputRfilename, outputSfilename, batchfilename=False, origScaling=True, cellScaledResults=False, logScaledResults=False, keepEffects=True, origRegress=False):
        self.kNN = kNN
        self.rawfilename = inrawfilename
        self.xtildeNeighborIndicesFilename = indexfilename
        self.xtildeNeighborDistancesFilename = distancefilename
        self.df_new_R = pd.DataFrame()
        self.df_new_S = pd.DataFrame()
        self.df_Rraw = pd.DataFrame()
        self.df_Regress_Pvalues = pd.DataFrame()
        self.outputRfilename = outputRfilename
        self.outputSfilename = outputSfilename
        self.batchfilename = batchfilename
        self.origScaling = origScaling if origScaling==False else True
        self.cellScaledResults = cellScaledResults if cellScaledResults==False else True
        self.logScaledResults = logScaledResults if logScaledResults==False else True
        self.keepEffects = keepEffects 
        self.origRegress = origRegress if origRegress==True else False

    def setOrigScaling(self,inOrigScale):
        self.origScaling = inOrigScale if inOrigScale==False else True

    def reportScaling(self):
        print('Scaling Original type is '+str(self.origScaling))
        print('LogScaledResults is '+str(self.logScaledResults))

    def getRmatrix(self):
        return (self.df_new_R)

    def getSmatrix(self):
        return (self.df_new_S)

    def reportParameters(self):
        print('Regress Data set')
        print('Input reads matrix '+self.rawfilename)
        print('Type of impute regression to use: origVersion ? :'+str(self.origRegress))
        print('Input Indices '+self.xtildeNeighborIndicesFilename)
        print('Input distances '+self.xtildeNeighborDistancesFilename)
        print('Output R filename '+self.outputRfilename)
        print('Output S filename '+self.outputSfilename)
     
    def regressCellValues(self, df_R, NNindex, cellIndexMap, datacelltotals):
        debug = False
        wlsStyleRegression=True
        #
        if (wlsStyleRegression):
            print('Perform regression with sample weighting and zero-zero point registration')
        timein = tm.time()
        RT = pd.DataFrame()  # TRansposed to allow efficient per-cell updates to be columns
        RTPvalues = pd.DataFrame() # Carry down P-values for double checking        
        icountCells=0
        numGenes = df_R.shape[1]
        #print('Truncated REGRESS list for testing')
        additionalMultiplier = 10.0 # 10*current max weight: Kirk wanted an additional weight applied to zero point
        numNNs = NNindex.shape[1] # NOTE this INCLUDES the target cell
        for icell in df_R.index.values:
            #print('icel is '+icell)
            NNlist = NNindex.loc[icell].tolist()[0:] # return integer IDs: Also keep the first entry which is redundant to cell name
            NNlistNames = cellIndexMap.loc[NNlist].values.flatten() # Cells for performing regressin
            if (len(NNlist) != len(NNlistNames)):
                print('mismatch: NNlist and NNListNames')
                print('cell maps')
                print(str(cellIndexMap))
                cellIndexMap.to_csv('cellMapIndex.tsv')
                print('detailed output')
                for ival in NNlist:
                    print('ival '+str(ival))
                    print('cell val '+str(cellIndexMap.loc[ival])+'XXX')
            NNlistReadData = df_R.loc[NNlistNames] # Get the actual current reads
            icellTotal = datacelltotals.loc[icell].values # For predicting new R for the target cell/gene pair
            Xdata = datacelltotals.loc[NNlistNames].values.flatten()
            Ydata = NNlistReadData.values
            newGeneValues = list()
            model = LinearRegression(n_jobs=8)
            icellTotal = datacelltotals.loc[icell].values
            if (wlsStyleRegression):
                Xdata = np.insert(Xdata, 0, 0.0)
                Yzeros = np.zeros(shape=(numGenes,1))  # Introduce a "fake" cell with total=0 and read=0 for each gene
                Ydata = np.insert(Ydata, 0, Yzeros.reshape(1,-1),axis=0) 
                sample_weights = [np.sqrt(x) for x in Xdata]  # Large valued Xs have better Ys
                sample_weights[0] = additionalMultiplier * np.max(sample_weights)
                model.fit(Xdata.reshape(-1,1), Ydata, sample_weights) 
            else:
                Xdata = np.insert(Xdata, 0, 0.0)
                Yzeros = np.zeros(shape=(numGenes,1)) 
                Ydata = np.insert(Ydata, 0, Yzeros.reshape(1,-1),axis=0)
                model.fit(Xdata.reshape(-1,1), Ydata)
            #
            # Could predict all K+1 cells at ones and update the input R matrix
            # What if any of the reads are negative ?
            YdataPredict = model.predict(icellTotal.reshape(-1,1)) # Expect only one column of data
            #print('predict '+str(YdataPredict))
            ydatamin = YdataPredict.min()
            if (ydatamin < 0.0):
                print('Negative fit new WLS with new weights found for '+str(YdataPredict.shape))
                zeroout_indices = np.argwhere(YdataPredict < 0.0)
                YdataPredict[ YdataPredict < 0.0] = 0.0
                #YdataPredict[zeroout_indices]=0.0
                #print(str(zeroout_indices))
                print('applied a builtin floor to cell fit '+str(zeroout_indices.shape)+' '+str(icell))
            # Some values can be < 0 probably because the cluster neighbors are poorly chosen
            # Should we set a floor of zero?
            RT[icell]=YdataPredict.flatten()
            RT.index = df_R.columns.values
        print('completed regressions '+str(tm.time()-timein))
        return (RT.T)

# NOTE NNLS is NOT parallel 
# ALSO NNLS cannot automatically handle regressing one X column onto many Y observations
# So to get this to work we must write our own parallel code to process all obs for a given cell total

    def regressCellValuesNNLS(self, df_R, NNindex, cellIndexMap, datacelltotals):
        """Here we DO NOT weight the regression terms. We DO NOT add the (0,0) point.
        We simply run the NNLS using defauilt minimization parameters
        """
        debug = False
        #
        timein = tm.time()
        RT = pd.DataFrame()  # Transposed to allow efficient per-cell updates to be columns
        RTPvalues = pd.DataFrame() # Carry down P-values for double checking        
        icountCells=0
        numGenes = df_R.shape[1]
        #print('Truncated REGRESS list for testing')
        numNNs = NNindex.shape[1] # NOTE this INCLUDES the target cell
        for icell in df_R.index.values:
            #print('icel is '+icell)
            NNlist = NNindex.loc[icell].tolist()[0:] # return integer IDs: Also keep the first entry which is redundant to cell name
            NNlistNames = cellIndexMap.loc[NNlist].values.flatten() # Cells for performing regressin
            if (len(NNlist) != len(NNlistNames)):
                print('NNLS mismatch: NNlist and NNListNames')
                print('cell maps')
                print(str(cellIndexMap))
                cellIndexMap.to_csv('cellMapIndex.tsv')
                print('detailed output')
                for ival in NNlist:
                    print('ival '+str(ival))
                    print('cell val '+str(cellIndexMap.loc[ival])+'XXX')
            NNlistReadData = df_R.loc[NNlistNames] # Get the actual current reads
            icellTotal = datacelltotals.loc[icell,'CELLTOTAL'] # For predicting new R for the target cell/gene pair
            Xdata = datacelltotals.loc[NNlistNames].values.flatten()
            Ydata = NNlistReadData.values
            newGeneValues = list()
            #Xdata = np.insert(Xdata, 0, 0.0)
            #Yzeros = np.zeros(shape=(numGenes,1)) 
            #Ydata = np.insert(Ydata, 0, Yzeros.reshape(1,-1),axis=0)
            w,rnorm = aggregateNNLS(Xdata.reshape(-1,1),Ydata)
            ###w,rnorm = nnls(Xdata.reshape(-1,1),Ydata))
            #
            # Could predict all K+1 cells at ones and update the input R matrix
            # What if any of the reads are negative ?
            YdataPredict = np.dot(icellTotal,w) # Expect only one column of data
            #print('predict '+str(YdataPredict))
            ydatamin = YdataPredict.min()
            if (ydatamin < 0.0):
                print('Negative fit NNLS found for '+str(YdataPredict.shape))
                zeroout_indices = np.argwhere(YdataPredict < 0.0)
                YdataPredict[ YdataPredict < 0.0] = 0.0
                #YdataPredict[zeroout_indices]=0.0
                #print(str(zeroout_indices))
                print('NNLS applied a builtin floor to cell fit: should never happen '+str(zeroout_indices.shape)+' '+str(icell))
            # Some values can be < 0 probably because the cluster neighbors are poorly chosen
            # Should we set a floor of zero?
            RT[icell]=YdataPredict.flatten()
            RT.index = df_R.columns.values
        print('completed NNLS regressions '+str(tm.time()-timein))
        return (RT.T)

    def writeNewR(self):
        self.df_new_R.to_csv(self.outputRfilename,sep=' ')
        print('wrote out new R matrix '+self.outputRfilename)
        print('keepEffect is '+str(self.keepEffects))

    def writeNewS(self):
        self.df_new_S.to_csv(self.outputSfilename,sep=' ')
        print('wrote out new S matrix '+self.outputSfilename)
        print('keepEffect is '+str(self.keepEffects))
 
    def getOutputRfilename(self):
        return (self.outputRfilename)

    def getOutputSfilename(self):
        return (self.outputSfilename)

# Another rwqs change from the drs giving rise to even uglier code

    def scaleResults(self, df_R, batchfile,origScaling,logScaledResults,cellScale):
        """ For the impute R (which includes group effects) scale 
        data for read counts, and then studentize the features to N(0,1)
        (value+1)*median-all-cell-read / call total read
        The post averaged cell totals is approx 2100 for a factor of 1
        """
        scaleBounds = pd.DataFrame()
        covs,oneHotTerms = getKnownGroupEffects(df_R,batchfile)
        print(str(df_R.columns))
        print(str(covs))
    
        df_R_X_i=df_R.drop(covs,axis=1,inplace=False)
        df_new_R = df_R[covs] # Hold back until the end
        df_R_X = df_R_X_i.astype(float) # Copy false or not doesn't update inplace
        factor=1.0
        total_reads = float(df_R_X.values.sum())
        num_reads = float(df_R_X.values.size)
        ave_total_reads = total_reads / num_reads
        # Sometimes the data come in as ints
        print('convert data type to float: use a mean method:')
        total_mean_c = pd.DataFrame(df_R_X.mean(axis=1)) # or do we want median values
        total_mean_c.columns=['TOTALC']
        total_correction_factor = factor * ave_total_reads / total_mean_c['TOTALC']
        if (cellScale):
            print('Initial Guess: Will also scale cel total counts')
        ###############################################
        if (cellScale):
            print('Preprocess cell scaling')
            if (logScaledResults):
                print('c+1 log scaling of counts')
                df_scale_new = np.log2((df_R_X+1.0).mul(total_correction_factor,axis=0))
            else:
                df_scale_new = df_R_X.mul(total_correction_factor,axis=0) # Scale cells to have same total reads
        else:
            df_scale_new = df_R_X
        # Lastly scale the features 
        # Getting lost when axis=1 and axis=0. Setting zero here gets us COLUMN scaling
        # Compute Max and Min values then store to disk for inverse scaling of final result
        ################################################
        # Dump min/max or mean/std FEATURE attributes to disk for subsequent inverse scaling
        if (origScaling):
            scaleBounds = pd.merge(df_scale_new.mean().to_frame(),df_new_scale.std().to_frame(),left_index=True,right_index=True)
            scaleBounds.columns = ['MEAN','STD']
        else:
            scaleBounds = pd.merge(df_scale_new.min().to_frame(),df_scale_new.max().to_frame(),left_index=True,right_index=True)
            scaleBounds.columns = ['MIN','MAX']
        ################################################
        if (origScaling):
            df_new = pd.DataFrame(sklearn.preprocessing.scale(df_scale_new, with_mean=True, with_std=True, copy=True,axis=0).tolist())
        else:
            print('Guess scaling: MinMax')
            scaler = MinMaxScaler(feature_range=(0.0,1.0)) # Might also want 0.1,0.9 instead
            df_new = pd.DataFrame(scaler.fit_transform(df_scale_new))
       # Now scale the cells
        df_new.index = df_R_X.index.values
        df_new.columns = df_R_X.columns.values

    # Add effects back into the data object

        finalS = pd.merge(df_new_R,df_new,left_index=True,right_index=True)
        df_new_S = finalS
        print('Scale the post-regress R function')
        print(str(df_new_S.shape))
        print('---------------')
        return(df_new_S)

    def scaleResultsPrevious(self,df_R,):
        """ For the impute R (which includes group effects) scale 
        data for read counts, and then studentize the features to N(0,1)
        (value+1)*median-all-cell-read / call total read
        The post averaged cell totals is approx 2100 for a factor of 1
        Perhaps do not wish to always take the log(v+1) of the new imputed results
        """
        factor = 10000.0
        # First ave number of reads per cell (or do we want median values?)
        total_reads = float(df_R.values.sum())
        num_reads = float(df_R.values.size)
        ave_total_reads = total_reads / num_reads
        # Alternative method use the median
        ## total_sum_c = pd.DataFrame(df_R.sum(axis=1))
        ## median_total_reads = total_sum_c.median()
        #
        ## total_median_c = pd.DataFrame(df_R.median(axis=1)) # or do we want median values
        ## total_median_c.columns=['TOTALC']
        total_mean_c = pd.DataFrame(df_R.mean(axis=1)) # or do we want median values
        total_mean_c.columns=['TOTALC']
        # Ave number of reads per cells divided by cell reads
        total_correction_factor = factor * ave_total_reads / total_mean_c['TOTALC']
        # If we feature scale after the cell scale, we might get cell totals of < 0
        if (self.origScaling):
           print('Mean center scaling')
           df_new_S = pd.DataFrame(sklearn.preprocessing.scale(df_R, with_mean=True, with_std=True, copy=True,axis=0).tolist())
        else:
           print('Regresss S Guess scaling: MinMax')
           scaler = MinMaxScaler(feature_range=(0.0,1.0)) # Might also want 0.1,0.9 instead
           df_new_S = pd.DataFrame(scaler.fit_transform(df_R))
        df_new_S.index = df_R.index.values
        df_new_S.columns = df_R.columns.values
        ##if (self.logScaledResults):
        ##    df_scale_R = np.log((df_R+1.0).mul(total_correction_factor,axis=0))
        ##else:
        ##    df_scale_R = df_R.mul(total_correction_factor,axis=0) # Scale cells to have same total reads
        # standardize FEATURES not cells
        print(str(df_new_S.shape))
        #print(str(df_new_R)) 
        print('---------------')
        return(df_new_S)

    def runProcess(self):

        #keepEffects=True

        print(' read matrix '+self.rawfilename)
        print(' Index neighbors '+self.xtildeNeighborIndicesFilename)
        print(' Distance neighbors '+self.xtildeNeighborDistancesFilename)
        t0=tm.time()
        RinData = pd.read_csv(self.rawfilename,delim_whitespace=True,index_col=0, header=0)
        NNindex = pd.read_csv(self.xtildeNeighborIndicesFilename,delim_whitespace=True,index_col=0,header=0)
        NNdistance = pd.read_csv(self.xtildeNeighborDistancesFilename,delim_whitespace=True,index_col=0,header=0)
        print(tm.time()-t0)
        # Fetch covariates to carry around separately
        columnEffects,oneHotList = getKnownGroupEffects(RinData,self.batchfilename)  # oneHotList not used at this time
        df_Z = RinData[columnEffects]
        df_Z.index = RinData.index.values
        df_R = RinData.drop(columnEffects,axis=1,inplace=False)
        ##df_R.set_index('CELL',inplace=True)
        # Total reads per CELL
        datacelltotals = pd.DataFrame(df_R.sum(axis=1))
        datacelltotals.reset_index(inplace=True)
        datacelltotals.columns=['CELL','CELLTOTAL']
        datacelltotals.set_index('CELL',inplace=True)
        # Integer cell ID to cell name
        cellIndexMap = buildCellIndexMap(NNindex)
        # Build regression
        print('Call regression')
        # Can this result in negative values for R?
        if (self.origRegress):
            df_new_R = self.regressCellValues(df_R, NNindex,cellIndexMap,datacelltotals)  # All Z columns have been removed
        else:
            df_new_R = self.regressCellValuesNNLS(df_R, NNindex,cellIndexMap,datacelltotals)
        print('min value of regressed R is '+str(df_new_R.values.min()))
        # Build the properly formatted output matrix including Z covariates
        numRcells = df_R.shape[0]
        numNNcells = NNindex.shape[0]
        if (numRcells != numNNcells):
            print('Cell rank of input R data and input K data are not equal. The resulting data object will take intersection')
            print('Need to investigate: should ABORT')
            print('R and NN cell ranks are '+str(numRcells)+' '+str(numNNcells))
            #sys.exit()
        print('R and NN cell ranks are '+str(numRcells)+' '+str(numNNcells))
        # Compute a feature scaled and log transformed version log the R matrix (called S)
        df_new_S = self.scaleResults(df_new_R, False, self.origScaling, self.logScaledResults, self.cellScaledResults)

        # To maintain computation using the so-called original sin model we need to muck
        # around with kepeping or not the effects: df_Z
        # Add effects back into the data object

        if (self.keepEffects):
            finalR = pd.merge(df_Z,df_new_R,left_index=True,right_index=True)
            finalS = pd.merge(df_Z,df_new_S,left_index=True,right_index=True)
        else:
            finalR = df_new_R
            finalS = df_new_S 
        self.df_new_R = finalR 
        self.df_new_S = finalS

        #print('Write to disk the new R matrix which is the imputed data and the S matrix which scales R as follows')
        #print('S = log( (R+1)*cellCorrectionFactor)')
        #self.writeNewR()
        #self.writeNewS()
        return ()



#
# Construct the data sets given the sizes n,p,k,and w
#
import time as tm
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from broonie.generate_neighbor_matrix import generateNeighborMatrixClass
from sklearn.externals.joblib import Parallel, delayed



#############################################################################
# inputfilename='60_MOUSEDATESEXTREATMENT_Xtilde_ZscoreOrder.tsv'
# No group effects are included in this file
# To get the parallelism working launch as a slurm job with 16 cores per node, and 1 node
# ask for 16 GB per core

knn = 10
njobs = 16
xfilename = './DATA/outputXtilde_0_60_Xtilde_ZscoreOrder.tsv'
inputXfilename = xfilename
outputIndices = 'dfIndices-test.tsv'
outputDistances = 'dfDistances-test.tsv'
print('Run clustering ' + inputXfilename)
y = generateNeighborMatrixClass(knn,inputXfilename,outputIndices,outputDistances,njobs)
y.reportParameters()
y.runProcess()
y.writeClusterInformation() # Write out the K and distance matrices
yindexfilename = y.getIndexFilename()
ydistancefilename = y.getDistanceFilename()

#################################### Third run the regression
print('clustering complete')




import time as tm
import pandas as pd
import numpy as np
from broonie.regress import regressDataClass

print('Begin regress')

inIndices = 'dfIndices.tsv'
inDistances = 'dfDistances.tsv'
inRdata = '/projects/sequence_analysis/vol1/prediction_work/CausalInference/CausalNetworking_forKirk/PCAProjectionImputationPipeline/MouseTumor_allMice_allGenes_Rawdata/mouse_raw_data_allcells_Notscaled.tsv'

knn=10
###################################################### group effect
# Straight up regression using the indices/distance and Raw data set. Default covariates are selected
print('Run regressions')

inputrawfilename = inRdata
yindexfilename = inIndices
ydistancefilename = inDistances
outputRfilename ='newRoutput-testing.tsv'
outputSfilename ='newSoutput-testing.tsv'
z = regressDataClass(knn,inputrawfilename,yindexfilename,ydistancefilename,outputRfilename,outputSfilename,origRegress=False)
z.reportParameters()
z.runProcess()
z.writeNewR()
z.writeNewS()

###################################################### group effect
# Use the filebased techniques for specifying possible effect names
# batchfile must contain ALL possible effects in the data set
batchfilename='listOfBatchEffects.txt'
inputrawfilename = inRdata
yindexfilename = inIndices
ydistancefilename = inDistances
outputRfilename ='newRoutput-testing3.tsv'
outputSfilename ='newSoutput-testing3.tsv'
z = regressDataClass(knn,inputrawfilename,yindexfilename,ydistancefilename,outputRfilename,outputSfilename,batchfilename,origScaling=True)
z.reportParameters()
z.reportScaling()
z.runProcess()
z.writeNewR()
z.writeNewS()

####################################################### USe Xtilde as input

inputrawfilename = 'testingOneXtildeMinMax_60_Xtilde_ZscoreOrder.tsv'
yindexfilename = inIndices
ydistancefilename = inDistances
outputRfilename ='newRoutputMinMax-testing.tsv'
outputSfilename ='newSoutputMinMax-testing.tsv'
z = regressDataClass(knn,inputrawfilename,yindexfilename,ydistancefilename,outputRfilename,outputSfilename,origScaling=False)
z.reportParameters()
z.reportScaling()
z.runProcess()
z.writeNewR()
z.writeNewS()




#################################################################
##
## These are the construction of the data from Ben using Zscores.
## So we do not filter by zeros. but we continue to use MAD() thresholding

## BUILD a new datafrae as follows. Columns CELLS MOUSE SEX DATE VEHICLE gene1 gene2 ....genep
## The gene values should be PRESCALED but not whitened. We can do thatlater
## The MOUSE value can be constructed from the CELL name

# MetaDataWithDate
##"nGene" "nUMI"  "Replicate"     "Treatment"     "Sex"   "Date"  "ID"

## 
## also The true Raw data has more cells than Ben's scaled data 
## SO we read in His f*ing daa here to provide a filter. (Damn guys !)

## So we have initially 16,600 Genes and 29,538 SingleCells
##
##################################################################

import time as tm
import pandas as pd
import numpy as np
import collections
import matplotlib.pyplot as plt
import math 
import scipy 
from scipy import stats
import sys
import sklearn.preprocessing

import sys
from numpy.random import seed
from numpy.random import randn
from statsmodels.graphics.gofplots import qqplot
import pylab
import scipy.stats as stats
from scipy import linalg

#from __future__ import print_function
#########################################################################

rootdir = '/projects/sequence_analysis/vol1/prediction_work/CausalInference/CausalNetworking_forKirk/'
rawfilename = rootdir+'rawResults/RawDataMatrix.txt'
metafilenameDate = rootdir+'rawResults/MetaDataMatrixWithDate.txt'

# Bens raw data 

benrootdir = '/projects/sequence_analysis/vol1/prediction_work/CausalInference/CausalNetworking_forKirk/'
benregressedfilename = rootdir+'regressedResults/ScaleDataMatrix.tsv'

t0=tm.time()
data = pd.read_csv(rawfilename,delim_whitespace=True)
print(tm.time()-t0)

regdata = pd.read_csv(benregressedfilename,delim_whitespace=True)

## Now exclude cells that ben excluded

listrawcolumns = data.columns.values
listregresscolumns = regdata.columns.values
listdiff = list(set(listrawcolumns)-set(listregresscolumns))

data.drop(listdiff,axis=1,inplace=True)

##
## Some basic plots of the raw Zscore imputed data
##

datameans = pd.DataFrame(data.mean(axis=1))
datameans.reset_index(inplace=True)
datameans.columns=['GENE','MEANS']
#datameans.plot(kind='line',x='GENE',y='MEANS',label='Zscore imputed data: Column MEANS of raw scRNAseq data')
#datameans.xlabel=('Gene')
#plt.ylabel=('Mean')
#plt.savefig('imputedData_GeneMeanCounts.pdf')
#plt.show()

##
## How about the cell mean reads?
##
datacellmeans = pd.DataFrame(data.mean())
datacellmeans.reset_index(inplace=True)
datacellmeans.columns=['CELL','CELLMEANS']
datacellmeans.sort_values(by=['CELL'],inplace=True)
#datacellmeans.plot(kind='line',x='CELL',y='CELLMEANS',label='Column MEANS of Zscore imputed scRNAseq data')
#datacellmeans.xlabel=('Cell')
#datacellmeans.ylabel=('CellMean')
#plt.savefig('imputedData_CellMeanCounts.pdf')
#plt.show()

##
## How about the cell mean reads?
##

datacellmeans = pd.DataFrame(data.mean())
datacellmeans.reset_index(inplace=True)
datacellmeans.columns=['CELL','CELLMEANS']
datacellmeans.sort_values(by=['CELL'],inplace=True)
#datacellmeans.plot(kind='line',x='CELL',y='CELLMEANS',label='Column MEANS of imputed scRNAseq data')
#datacellmeans.xlabel=('Cell')
#datacellmeans.ylabel=('CellMean')
#plt.savefig('rawData_CellMeanCounts.pdf')
#plt.show()

######################################################################
## Compute gene MAD() values then keep 5000 top
## CHeck the profile to see how many to keep

datamad = pd.DataFrame(data.mad(axis=1))
datamad.reset_index(inplace=True)
datamad.columns=['GENE','MAD']
datamad.sort_values(by='MAD',ascending=False,axis=0,inplace=True)
#datamad.plot(kind='line',x='GENE',y='MAD',label='Kirk: Imputed data: MAD() of imputed scRNAseq data')
#plt.ylabel=('MAD')
#plt.savefig('imputedData_GeneMADCounts.pdf')
#plt.show()

## Sort data set  and include only the4 Top 5000 MAD values
topGenes = datamad.head(5000)
topGenesList = topGenes['GENE'].tolist()
pd.DataFrame(topGenesList).to_csv('topGenes_MAD.tsv')

# Kirk wants to keep all genes for now

######################################################################
## Get list of TUMOR names for future use and proportions
## Filter-keep only Vehicle cells

def tumor_within_cellName(tumorname,cellname):
    return tumorname in cellname

def computeSizeFactor_cellname(cellname, counterObject):
    name = cellname.split('_')[0]
    numbcells = float(counterObject[name])
    total = float(sum(counterObject.values()))
    return numbcells/total

#fulllistNames = list(data.loc[topGenesList].columns.values)
fulllistNames = list(data.columns.values)
unique_tumors = set([ x.split('_')[0] for x in fulllistNames ])
#{'MB10', 'MB13', 'MB05', 'MB14', 'MB06', 'MB04', 'MB15', 'MB07', 'MB12', 'MB09'}
allentries_tumors_temp = list([ x.split('_')[0] for x in fulllistNames ])
from collections import Counter
counter = collections.Counter(allentries_tumors_temp)

## 1) Get some base data: Compute total ave read count from the Raw data
## just for comparisons
## We assume missingness is not to0 bad. But it seems we should QC the raw data 
## to get the following values  (?)

## skip the scaling of the reads

#data_new = data.loc[topGenesList]
data_new = data

# Confirm total reads per cell are the same
cellsums = data_new.sum(axis=0)

# What about total reads per gene
genesums = data_new.sum(axis=1)
#genesums.hist(bins=20,range=[0,100000])
#plt.savefig('imputedData_GeneSums_postScale.pdf')
#plt.show()

###############################################################################3
# No exclusons based on zeros

## Post QC Plot information
# Per gene
datameans = pd.DataFrame(data_new.mean(axis=1))
datameans.reset_index(inplace=True)
datameans.columns=['GENE','VARIANCE']
#datameans.plot(kind='line',x='GENE',y='VARIANCE',label='Imputed data: Column Scaled-MEANS of imputed scRNAseq data')
#datameans.xlabel=('Gene')
#datameans.ylabel=('Scaled-MEANS')
#plt.savefig('imputedData_GeneMeanCounts_removeMissingness.pdf')
#plt.show()

# Per Cell
datacellmeans = pd.DataFrame(data_new.mean())
datacellmeans.reset_index(inplace=True)
datacellmeans.columns=['CELL','CELLMEANS']
datacellmeans.sort_values(by=['CELL'],inplace=True)
#datacellmeans.plot(kind='line',x='CELL',y='CELLMEANS',label='Column MEANS of imputed scRNAseq data')
#datacellmeans.xlabel=('Cell')
#datacellmeans.ylabel=('CellMean')
#plt.savefig('rawData_CellMeanCounts_removeMissingness.pdf')
#plt.show()

############################################################################################
# cell reads for all mice have now been QC'd and scaled. We start filtering away non-vehicle, non-tumor data
# Then build data into the dataframe to apture some batch effects such as Date and Sex

## First Remove all CELLs that are not actually part of a Tumor using the MAP from Kirk
## /proj/kirklab/users/KirkFiles/
## We also only want VEHICLE cells

### This whoe procedure has changed AGAIN

# NODENAMES THAT ARE NOT TUMORS: Remaining terms have the nomenclaturee NODE A, NODE B, etc
# Dec 12. Per kirk. Keep al cell types but exclude the vismodigib mice: onlty keep vehicle

#############################################################################################
# PLACE A DUMMY TERM INTO NOTTUMORS AND RERUN

#notTumors =['Endothelial Cells','Vascular Fibroblasts','Astrocytes','Endothelial','Microglia','Neurons','Oligodendrocytes','Vascular']
notTumors =['TEST']
metadata = pd.read_csv(metafilenameDate,delim_whitespace=True)

# Get list of cells that are NOT tumors
listNotTumorCells_df = metadata.ix[((metadata['ID'].isin(notTumors)))]
listNotTumorCells = list(listNotTumorCells_df.index.values)

#############################################################################################
## Now get list of cells. Keep all of 'em

listNotVehicleCells_df = metadata.ix[ metadata['Treatment'] == 'DUMMY']
listNotVehicleCdells = list(listNotVehicleCells_df.index.values)


# Union unique of exclude lists
excludeList = set(listNotTumorCells+listNotVehicleCdells)
data_new_tumor = data_new.copy()
data_new_tumor.drop(excludeList,axis=1,inplace=True)

cellNames = data_new_tumor.iloc[1].to_frame()
cellNamesList = cellNames.index.tolist()
geneNames = data_new_tumor.iloc[:,1].to_frame()
geneNamesList = geneNames.index.tolist()
geneNamesList_cell = geneNamesList

# get data from numpy and convert to df
# Keep only the top5000
# temp = data_new_tumor.T[topGenesList].copy()
# data_new_tumor = temp.T
 
# NO SCALING

data_new_tumor_scaled_T = data_new_tumor.T
data_new_tumor_scaled_T['CELL'] = cellNamesList
data_new_tumor_scaled_T.set_index('CELL',inplace=True)
data_new_tumor_scaled_T.columns = geneNamesList_cell

# Add metadata to the scaled data object (SEX,DATE,MOUSE)

scaledData = data_new_tumor_scaled_T.copy()
newcolumn = metadata.loc[scaledData.index.values][['Treatment','Sex','Date','ID']]

##
## Splitting on spaces wil cause problems. Simply remove spaces and let downstream live with it
##newcolumn.ID = newcolumn.ID.str.split(' ').str.get(1)
newcolumn.ID = [x.replace(' ','') for x in list(newcolumn.ID)]

# Now let's build a column that only has a mouse ID. We can convert to multiple rows of 
# indicator variables at some later time

mouseIndex = [ int(x.split('_')[0][2:]) for x in newcolumn.index ]
newcolumn['MOUSE']=mouseIndex

# Build final dataframe for disk storage
combinedScaledData = pd.merge(newcolumn,scaledData,left_index=True,right_index=True)
combinedScaledData.rename(columns={'Treatment':'TREATMENT','Sex':'SEX','Date':'DATE','ID':'NODEID'},inplace=True)

outfilename = 'mouse_raw_data_allcells_Notscaled.tsv'
combinedScaledData.to_csv(outfilename,header=True,sep=" ",index=True)

print('raw data processing completed')




#!/usr/bin/env python
from __future__ import division, print_function
from setuptools import setup, find_packages

import os
import sys
import subprocess
import textwrap

def readme():
    with open('README.md') as f:
        return f.read()

def configuration(parent_package='',top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('broonie', parent_package, top_path)
    config.add_subpackage('covariate_projection_reduction')
    config.add_subpackage('mergeMetadataToData')    
    config.make_config_py() 
    return config

def setup_package():
    src_path = os.path.dirname(os.path.abspath(sys.argv[0]))
    old_path = os.getcwd()
    os.chdir(src_path)
    sys.path.insert(0, src_path)
    
    metadata = dict(name='broonie',
                    version='0.1.7',
                    description='SVD PCA Projection Imputation Pipeline',
                    long_description=readme(),
                    classifiers=[
                        'Development Status :: 3 - Alpha',
                        'Intended Audience :: Science/Research',
                        'Intended Audience :: Developers',
                        'Programming Language :: Python',
                        'Programming Language :: Python :: 3',
                        'Topic :: Software Development',
                        'Topic :: Scientific/Engineering',
                        'Operating System :: Unix',
                        'Operating System :: POSIX :: Linux'                        
                    ],
                    keywords='pca projection imputation pipeline batch corrected',
                    url='http://github.com/RENCI/broonie',
                    author='Jeffery Tilson',
                    author_email='jtilson@renci.org',
                    packages=find_packages(),
                    platforms = [ "Windows", "Linux", "Solaris", "Mac OS-X", "Unix" ],
                    install_requires=['markdown', 'pandas', 'numpy', 'matplotlib', 'scikit-learn', 'scipy'],
                    test_suite='nose.collector',
                    zip_safe=False)

    setup(**metadata)
    return

if __name__ == '__main__':
    setup_package();



#
# Construct the data sets given the sizes n,p,k,and w
#
import time as tm
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from broonie.project_group import projectGroupDataClass
from broonie.generate_neighbor_matrix import generateNeighborMatrixClass
from broonie.regress import regressDataClass
from broonie.construct_initial_guess import constructInitialGuess
import sys
from sklearn.externals.joblib import Parallel, delayed

#############################################################################
# Define functions for convergence testing
# X matrices include all effects: So we want to remove them all (again)

maxDiffElemsThreshold = 0.1
maxFrobeniusThreshold = 0.05
maxIterations = 12

iterationMode = 'raw'
##iterationMode = 'incremental'

def dropAllEffects(df_X):
    currentPossibleColumnEffects = ['TREATMENT','MOUSE','SEX','DATE','NODEID']
    allColumns = df_X.columns.values
    foundColumns = list(set(allColumns) & set(currentPossibleColumnEffects))
    print('Number of check convergence: group effect columns in data is '+str(foundColumns))
    print('They are: '+str(foundColumns))
    return(df_X.drop(foundColumns,axis=1,inplace=False))

def checkConvergence(df_inX, df_inXold):
    """It has happened that the two input matrices have different row orders
    So we must address that here
    We do not anticipate checking Xtilde matrices. Rather R or S. Thus we must
    account for the possible covariates
    """
    print('in 1 '+str(df_inX.shape))
    print('in 2 '+str(df_inXold.shape))
    index1 = df_inX.index.values
    index2 = df_inXold.index.values
    if (index1.shape[0] != index2.shape[0]):
        print('different numbers of cells in the checkConvrgence: Abort')
        print(str(df_inX.shape))
        print(str(df_inXold.shape))
        sys.exit()
    combined = list(set(index1).intersection(index2))

    # first do they contain the same elements?

    converged = False
    df_X = dropAllEffects(df_inX.loc[combined]) 
    df_Xold = dropAllEffects(df_inXold.loc[combined])
    maxDiffElems =  (df_X-df_Xold).values.max()
    print('Max element '+str(maxDiffElems))
    if (maxDiffElems <= maxDiffElemsThreshold):
        print('X coverged on a maximum per-element threshold '+str(maxDiffElems))
        converged=True
    print('Max diff elem '+str(maxDiffElems))
    # Compute the frobenious norm  based convergence criteria
    print('try froben norm ')
    frac = np.linalg.norm(df_X-df_Xold) / np.linalg.norm(df_X)
    print('Convergence frac is '+str(frac))
    if (frac <= maxFrobeniusThreshold):
        print('X coverged on a frobenius threshold '+str(frac))
        converged=True
    print('Frobenious norm '+str(frac))
    return (converged)

print('Begin imputation')

# Number of parallel jobs for the cluster code
# Presumably threadbased. So don.t go too big. 8 or 16 at the most
# at njobs=1 the jobs takes at ~15,000 secs and at njobs=16 total time is 1,903 sec.
njobs = 16 
# Specify subspace for group effect reduction
k = 60
# Specify number of NNs for regression (actually we use knn+1 becasue we include the target cell
knn = 10

# Initial Read data (R naught)

# RAW DATA
rawfilename='/projects/sequence_analysis/vol1/prediction_work/CausalInference/CausalNetworking_forKirk/PCAProjectionImputationPipeline/MouseTumor_allMice_allGenes_Rawdata/mouse_raw_data_allcells_Notscaled.tsv'

# Group effects to exclude at every iteration
inlist='MOUSE_DATE_SEX_TREATMENT'

# Fetch an initial guess. In this example we simply take the RAW data and also Read Ben's associzated "regressed" data.
# Bens data is the seurat processed scaling of the Raw data. Moving forward, we may want to process initial guesses differently


outfilename = 'initialGuess.tsv'
initialGuess = constructInitialGuess(rawfilename,outfilename,batchfile=False,origScaling=False,logScaledResults=False)
initialGuess.setOrigScaling(False)
initialGuess.reportScaling()
#initialGuess.runInitialGuess()
initialGuess.fetchInitialGuess() # Take Seurat imputed data rescale to MinMax (0,1) and proceed
initialGuessFilename = initialGuess.getOutputFilename() # return Ben's seurat regressed data set as initial guess
print('output bounds filename is '+initialGuess.getBoundsFilename())
scaleBounds = initialGuess.fetchScaleBounds()


# Initialize the iterative procedure 
inputfilename = initialGuessFilename
inputregressionfilename = rawfilename

# Try using the pregressed file as the "raw" data
# rawfilename = inputfilename

istart = 0 
kindicesprevious = None
converged=False

print('Read RAW input file '+rawfilename)
RinData = pd.read_csv(rawfilename,index_col=0,delim_whitespace=True) # Bootstrap for the convergence testing
Sold = None 
Scurrent = RinData

if (iterationMode == 'raw'):
    print('Iteration will impute always versus RAW R')
else:
    print('Iteration will impute always versus current R')

while istart < maxIterations:
    #
    # if scaling by cells and features neede then do it here> Not required for initialGuess data
    print('Start the next new iteration'+str(istart))
###################################################### group effect
    outtildemeta='outputXtilde_'+str(istart)+'_'
    print('Run projection')
    x = projectGroupDataClass(k,inputfilename,inlist,outtildemeta,scale=True) # regress studenizes already
    x.reportParameters()
    x.runProjection()
    xtilde = x.fetchXtilde() # Keep this for testing convergence
    xfilename = x.getXtildeFilename()
###################################################### check convergence
# Note rows might be differently ordered
# Need to double check the test
# Check convergece of the R matrix ( One could also use the S matrix)
    if (Sold is not None):
        converged = checkConvergence(Scurrent, Sold)
    print('Convergence testing status is '+str(converged))
    if (converged):
        print('S has converged')
        break
##################################### Second. reclustering
# inputfilename='60_MOUSEDATESEXTREATMENT_Xtilde_ZscoreOrder.tsv'
# No group effects are included in this file
    inputXfilename = xfilename
    outputIndices = 'dfIndices-'+str(istart)+'.tsv'
    outputDistances = 'dfDistances-'+str(istart)+'.tsv'
    print('Run clustering ' + inputXfilename)
    y = generateNeighborMatrixClass(knn,inputXfilename,outputIndices,outputDistances,njobs)
    y.reportParameters()
    y.runProcess()
    y.writeClusterInformation() # Write out the K and distance matrices
    yindexfilename = y.getIndexFilename()
    ydistancefilename = y.getDistanceFilename()
    kindices = y.reportIndices()
#################################### Third run the regression
    print('Run regressions')
    inputrawfilename = rawfilename
    zindexfilename = yindexfilename
    zdistancefilename = ydistancefilename
    outputRfilename ='newRoutput-'+str(istart)+'.tsv'
    outputSfilename ='newSoutput-'+str(istart)+'.tsv'
    Sold = Scurrent
    print(inputrawfilename)
    print(yindexfilename)
    print(ydistancefilename)
    #
    #use the oldS instead of raw R inputrawfilename
    #Inputfilename SHOULD not contain negative terms
    z = regressDataClass(knn,inputregressionfilename,yindexfilename,ydistancefilename,outputRfilename,outputSfilename)
    z.reportParameters()
    z.runProcess()
    z.writeNewR()
    z.writeNewS()
    #inputrawfilename = z.getOutputRfilename() # This still contains effects like the rawR
    #############################################################
    #
    xtildeprevious = xtilde # Hold for next convergence test
    # This approach results in osscilation of the Max elements
    Scurrent = z.getSmatrix()
    #inputRmatrix = z.getRawRmatrix()
    inputfilename = z.getOutputSfilename() # Pass back to the OG methods
    if (iterationMode == 'raw'):
        inputregressionfilename = rawfilename
    else:
        print('Iteration will impute always versus RAW R')
        inputregressionfilename = z.getOutputRfilename()
    #
    istart += 1
#
if (converged):
    print('Iterations has converged')
print('Exit the loop with number of iteration = '+str(istart))
print('imputation completed')
print('Final imputted file is '+ z.getOutputRfilename())

